#
# Copyright (c) 2013-2020, Dell Inc. or its subsidiaries.
# All rights reserved.
# See file LICENSE for licensing information.
#
# Module Name:
#
#        durable.py
#
# Abstract:
#
#        Durable handle tests
#
# Authors: Arlene Berry (arlene.berry@emc.com)
#

from builtins import map

import array
import random
import time

import pike.model
import pike.smb2
import pike.test
import pike.ntstatus


# for buffer too small
class InvalidNetworkResiliencyRequestRequest(pike.smb2.NetworkResiliencyRequestRequest):
    def _encode(self, cur):
        cur.encode_uint32le(self.timeout)
        cur.encode_uint16le(self.reserved)


@pike.test.RequireCapabilities(pike.smb2.SMB2_GLOBAL_CAP_LEASING)
class DurableHandleTest(pike.test.PikeTest):
    share_all = (
        pike.smb2.FILE_SHARE_READ
        | pike.smb2.FILE_SHARE_WRITE
        | pike.smb2.FILE_SHARE_DELETE
    )
    lease1 = array.array("B", map(random.randint, [0] * 16, [255] * 16))
    lease2 = array.array("B", map(random.randint, [0] * 16, [255] * 16))
    r = pike.smb2.SMB2_LEASE_READ_CACHING
    rw = r | pike.smb2.SMB2_LEASE_WRITE_CACHING
    rh = r | pike.smb2.SMB2_LEASE_HANDLE_CACHING
    rwh = rw | rh
    buffer_too_small_error = pike.ntstatus.STATUS_INVALID_PARAMETER

    def create(
        self,
        chan,
        tree,
        durable,
        lease=rwh,
        lease_key=lease1,
        disposition=pike.smb2.FILE_SUPERSEDE,
    ):
        return chan.create(
            tree,
            "durable.txt",
            access=pike.smb2.FILE_READ_DATA
            | pike.smb2.FILE_WRITE_DATA
            | pike.smb2.DELETE,
            share=self.share_all,
            disposition=disposition,
            oplock_level=pike.smb2.SMB2_OPLOCK_LEVEL_LEASE,
            lease_key=lease_key,
            lease_state=lease,
            durable=durable,
        ).result()

    def durable_test(self, durable):
        chan, tree = self.tree_connect()

        handle1 = self.create(chan, tree, durable=durable)

        self.assertEqual(handle1.lease.lease_state, self.rwh)
        self.assertTrue(handle1.is_durable)

        chan.close(handle1)

    def durable_reconnect_test(self, durable):
        chan, tree = self.tree_connect()

        handle1 = self.create(chan, tree, durable=durable)

        self.assertEqual(handle1.lease.lease_state, self.rwh)
        self.assertTrue(handle1.is_durable)

        # Close the connection
        chan.connection.close()

        chan2, tree2 = self.tree_connect()

        # Request reconnect
        handle2 = self.create(chan2, tree2, durable=handle1)
        self.assertEqual(handle2.lease.lease_state, self.rwh)
        self.assertTrue(handle2.is_durable)
        chan2.close(handle2)

    def durable_reconnect_fails_client_guid_test(self, durable):
        chan, tree = self.tree_connect()

        handle1 = self.create(chan, tree, durable=durable)
        self.assertTrue(handle1.is_durable)

        self.assertEqual(handle1.lease.lease_state, self.rwh)

        # Close the connection
        chan.connection.close()

        chan2, tree2 = self.tree_connect(client=pike.model.Client())

        with self.assert_error(pike.ntstatus.STATUS_OBJECT_NAME_NOT_FOUND):
            handle2 = self.create(chan2, tree2, durable=handle1)

        chan2.connection.close()

        chan3, tree3 = self.tree_connect()

        handle3 = self.create(chan3, tree3, durable=handle1)
        self.assertTrue(handle3.is_durable)

        chan3.close(handle3)

    def durable_invalidate_test(self, durable):
        chan, tree = self.tree_connect()

        handle1 = self.create(chan, tree, durable=durable)
        self.assertEqual(handle1.lease.lease_state, self.rwh)
        self.assertTrue(handle1.is_durable)

        # Close the connection
        chan.connection.close()

        chan2, tree2 = self.tree_connect(client=pike.model.Client())

        # Invalidate handle from separate client
        handle2 = self.create(
            chan2,
            tree2,
            durable=durable,
            lease=self.rw,
            lease_key=self.lease2,
            disposition=pike.smb2.FILE_OPEN,
        )
        self.assertEqual(handle2.lease.lease_state, self.rw)
        chan2.close(handle2)

        chan2.connection.close()

        chan3, tree3 = self.tree_connect()

        # Reconnect should now fail
        with self.assert_error(pike.ntstatus.STATUS_OBJECT_NAME_NOT_FOUND):
            handle3 = self.create(chan3, tree3, durable=handle1)

    @pike.test.RequireDialect(pike.smb2.DIALECT_SMB2_1)
    def test_resiliency_reconnect_before_timeout(self, durable=True):
        chan, tree = self.tree_connect()
        handle1 = self.create(chan, tree, durable=durable)
        self.assertEqual(handle1.lease.lease_state, self.rwh)
        self.assertEqual(handle1.is_durable, durable)
        timeout = 100

        a = chan.network_resiliency_request(handle1, timeout=timeout)
        self.assertTrue(handle1.is_resilient)

        # Close the connection
        chan.connection.close()
        time.sleep((timeout - 10) / 1000.0)

        chan2, tree2 = self.tree_connect(client=pike.model.Client())

        handle2 = self.create(
            chan2,
            tree2,
            durable=durable,
            lease=self.rw,
            lease_key=self.lease2,
            disposition=pike.smb2.FILE_OPEN,
        )
        # resiliency interact handle2's lease_status(before rw, now r)
        self.assertEqual(handle2.lease.lease_state, self.r)
        chan2.close(handle2)
        chan2.connection.close()
        chan3, tree3 = self.tree_connect()
        handle3 = self.create(chan3, tree3, durable=handle1)
        self.assertTrue(handle3.is_resilient)

        # cause of the resiliency, handle1.fileid would be equals to handle3's.
        self.assertEqual(handle1.file_id, handle3.file_id)
        chan3.close(handle3)

    @pike.test.RequireDialect(pike.smb2.DIALECT_SMB2_1)
    def test_resiliency_reconnect_after_timeout(self, durable=True):
        chan, tree = self.tree_connect()
        handle1 = self.create(chan, tree, durable=durable, lease=self.rwh)
        self.assertEqual(handle1.lease.lease_state, self.rwh)
        self.assertEqual(handle1.is_durable, durable)
        timeout = 100
        a = chan.network_resiliency_request(handle1, timeout=timeout)
        self.assertTrue(handle1.is_resilient)

        # Close the connection
        chan.connection.close()
        time.sleep(timeout / 1000.0 + 5.0)  # timeout

        chan2, tree2 = self.tree_connect(client=pike.model.Client())
        # Invalidate handle from separate client
        handle2 = self.create(
            chan2,
            tree2,
            durable=durable,
            lease=self.rw,
            lease_key=self.lease2,
            disposition=pike.smb2.FILE_OPEN,
        )

        self.assertEqual(handle2.lease.lease_state, self.rw)
        chan2.close(handle2)
        chan2.connection.close()
        chan3, tree3 = self.tree_connect()

        # Reconnect should fail(resiliency timeout)
        with self.assert_error(pike.ntstatus.STATUS_OBJECT_NAME_NOT_FOUND):
            handle3 = self.create(chan3, tree3, durable=handle1)

    @pike.test.RequireDialect(pike.smb2.DIALECT_SMB2_1)
    def test_resiliency_buffer_too_small(self, durable=True):
        chan, tree = self.tree_connect()
        handle1 = self.create(chan, tree, durable=durable)
        self.assertEqual(handle1.is_durable, durable)

        timeout = 5
        with self.assert_error(self.buffer_too_small_error):
            smb_req = chan.request(obj=handle1.tree)
            ioctl_req = pike.smb2.IoctlRequest(smb_req)
            # replace with invalid request that has short input buffer
            invalid_nrr_req = InvalidNetworkResiliencyRequestRequest(ioctl_req)
            ioctl_req.file_id = handle1.file_id
            ioctl_req.flags = pike.smb2.SMB2_0_IOCTL_IS_FSCTL
            invalid_nrr_req.timeout = timeout
            invalid_nrr_req.reserved = 0
            a = chan.connection.transceive(ioctl_req.parent.parent)[0]

        self.assertEqual(handle1.lease.lease_state, self.rwh)
        chan.connection.close()

        chan2, tree2 = self.tree_connect(client=pike.model.Client())
        handle2 = self.create(
            chan2,
            tree2,
            durable=durable,
            lease=self.rw,
            lease_key=self.lease2,
            disposition=pike.smb2.FILE_OPEN,
        )

        self.assertEqual(handle2.lease.lease_state, self.rw)
        chan2.close(handle2)
        chan2.connection.close()

        chan3, tree3 = self.tree_connect()
        # buffer too small
        with self.assert_error(pike.ntstatus.STATUS_OBJECT_NAME_NOT_FOUND):
            handle3 = self.create(chan3, tree3, durable=handle1)

    # the resiliency_upgrade tests start with a non-durable handle
    # and upgrade to resilient handle by use of ioctl
    @pike.test.RequireDialect(pike.smb2.DIALECT_SMB2_1)
    def test_resiliency_upgrade_reconnect_before_timeout(self):
        self.test_resiliency_reconnect_before_timeout(durable=False)

    @pike.test.RequireDialect(pike.smb2.DIALECT_SMB2_1)
    def test_resiliency_upgrade_reconnect_after_timeout(self):
        self.test_resiliency_reconnect_after_timeout(durable=False)

    @pike.test.RequireDialect(pike.smb2.DIALECT_SMB2_1)
    def test_resiliency_upgrade_buffer_too_small(self):
        self.test_resiliency_buffer_too_small(durable=False)

    # Request a durable handle
    @pike.test.RequireDialect(pike.smb2.DIALECT_SMB2_1)
    def test_durable(self):
        self.durable_test(True)

    # Reconnect a durable handle after a TCP disconnect
    @pike.test.RequireDialect(pike.smb2.DIALECT_SMB2_1)
    def test_durable_reconnect(self):
        self.durable_reconnect_test(True)

    # Reconnecting a durable handle after a TCP disconnect
    # fails with STATUS_OBJECT_NAME_NOT_FOUND if the client
    # guid does not match
    @pike.test.RequireDialect(pike.smb2.DIALECT_SMB2_1)
    def test_durable_reconnect_fails_client_guid(self):
        self.durable_reconnect_fails_client_guid_test(True)

    # Breaking the lease of a disconnected durable handle
    # (without HANDLE) invalidates it, so a subsequent
    # reconnect will fail.
    @pike.test.RequireDialect(pike.smb2.DIALECT_SMB2_1)
    def test_durable_invalidate(self):
        self.durable_invalidate_test(True)

    # Request a durable handle via V2 context structure
    @pike.test.RequireDialect(pike.smb2.DIALECT_SMB3_0)
    def test_durable_v2(self):
        self.durable_test(0)

    # Reconnect a durable handle via V2 context structure
    @pike.test.RequireDialect(pike.smb2.DIALECT_SMB3_0)
    def test_durable_reconnect_v2(self):
        self.durable_reconnect_test(0)

    # Reconnecting a durable handle (v2) after a TCP disconnect
    # fails with STATUS_OBJECT_NAME_NOT_FOUND if the client
    # guid does not match
    @pike.test.RequireDialect(pike.smb2.DIALECT_SMB3_0)
    def test_durable_reconnect_v2_fails_client_guid(self):
        self.durable_reconnect_fails_client_guid_test(0)

    # Breaking the lease of a disconnected durable handle v2
    # (without HANDLE) invalidates it, so a subsequent
    # reconnect will fail.
    @pike.test.RequireDialect(pike.smb2.DIALECT_SMB3_0)
    def test_durable_v2_invalidate(self):
        self.durable_invalidate_test(0)



#
# Copyright (c) 2020, Dell Inc. or its subsidiaries.
# All rights reserved.
# See file LICENSE for licensing information.
#
"""
Pike/pytest integration

Available markers:

    - ``require_dialect(min_dialect, max_dialect)``: raise a
      :py:class:`~pike.test.DialectMissing` exception if the target server does
      not support a dialect in the given range. This exception will be raised
      after Negotiate.
    - ``require_capability(global_capabilities)``: raise a
      :py:class:`~pike.test.CapabilityMissing` exception if the target server
      does not support all of the OR'd global capability flags.  This exception
      will be raised after Negotiate.
    - ``require_share_capability(share_capabilities)``: raise a
      :py:class:`~pike.test.ShareCapabilityMissing` exception if the target
      share does not support all of the OR'd share capability flags.  This
      exception will be raised after Tree Connect.
"""

import uuid
import time

import pytest

import pike.test


# pytest API
def pytest_configure(config):
    """
    register the marks so that they show up in `pytest --markers` as well as
    not raising errors in strict marker mode
    """
    config.addinivalue_line(
        "markers",
        (
            "require_dialect(min_dialect, max_dialect): "
            "when using the TreeConnect fixture, if the remote server "
            "does not advertise a dialect within the given range "
            "then skip the test."
        ),
    )
    config.addinivalue_line(
        "markers",
        (
            "require_capabilities(global_capabilities): "
            "when using the TreeConnect fixture, if the remote server "
            "does not advertise support for all capabilities "
            "then skip the test."
        ),
    )
    config.addinivalue_line(
        "markers",
        (
            "require_share_capabilities(share_capabilities): "
            "when using the TreeConnect fixture, if the remote share "
            "does not advertise support for all share capabilities "
            "then skip the test."
        ),
    )


def get_mark_value(item, mark_name, default):
    for mark in item.iter_markers(name=mark_name):
        if len(mark.args) > 1:
            return mark.args
        return mark.args[0]
    return default


@pytest.fixture
def pike_TreeConnect(request):
    """
    Function-scope fixture returns a wrapper function around
    :py:class:`pike.TreeConnect`.

    The wrapper-function must be called to establish the connection and the
    returned object may be used as a contextmanager.

    The wrapper-function will pass through any ``require_dialect``,
    ``require_capabilities``, and ``require_share_capabilities`` marks on the
    test item and raise an exception if the target server doesn't support the
    required values.

    :return: Callable[..., pike.TreeConnect]
    """
    marks = (
        ("require_dialect", (0, float("inf"))),
        ("require_capabilities", 0),
        ("require_share_capabilities", 0),
    )

    def TreeConnect(*args, **kwargs):
        # update requirements from markers
        for mark_name, default_value in marks:
            kwargs[mark_name] = kwargs.get(
                mark_name,
                get_mark_value(request.node, mark_name, default_value),
            )
        # special case dialect range (must be a 2-tuple)
        if not isinstance(kwargs["require_dialect"], tuple):
            kwargs["require_dialect"] = (kwargs["require_dialect"], float("inf"))
        return pike.test.TreeConnect(*args, **kwargs)

    # carry the docstring to the wrapper function
    TreeConnect.__doc__ = pike.TreeConnect.__doc__

    return TreeConnect


@pytest.fixture
def pike_tmp_path(pike_TreeConnect):
    """
    Function-scope fixture reaturns a :py:class:`~pike.path.PikePath` rooted at
    a timestamped subdirectory of the share.

    The temporary directory is not deleted during tearDown and should be
    removed by the test case itself.

    A function's :py:func:`pike_TreeConnect` fixture will access the same
    share, however, via a separate SMB2 session.
    """
    with pike_TreeConnect() as tc:
        test_root = tc / "pike_{}_{}".format(
            time.strftime("%Y-%m-%d_%H%M%S"), str(uuid.uuid4())[:5]
        )
        test_root.mkdir()
        yield test_root



#
# Copyright (c) 2013-2020, Dell Inc. or its subsidiaries.
# All rights reserved.
# See file LICENSE for licensing information.
#
# Module Name:
#
#        persistent.py
#
# Abstract:
#
#        Persistent handle tests
#
# Authors: Paul Martin (paul.o.martin@emc.com)
#

from builtins import map

import array
import random
import time

import pike.model
import pike.smb2 as smb2
import pike.test as test
import pike.ntstatus as ntstatus

# Constants
LEASE_R = smb2.SMB2_LEASE_READ_CACHING
LEASE_RW = LEASE_R | smb2.SMB2_LEASE_WRITE_CACHING
LEASE_RH = LEASE_R | smb2.SMB2_LEASE_HANDLE_CACHING
LEASE_RWH = LEASE_RW | LEASE_RH
SHARE_ALL = smb2.FILE_SHARE_READ | smb2.FILE_SHARE_WRITE | smb2.FILE_SHARE_DELETE


class InvalidNetworkResiliencyRequestRequest(pike.smb2.NetworkResiliencyRequestRequest):
    def _encode(self, cur):
        cur.encode_uint32le(self.timeout)
        cur.encode_uint16le(self.reserved)


@test.RequireDialect(smb2.DIALECT_SMB3_0)
@test.RequireCapabilities(smb2.SMB2_GLOBAL_CAP_PERSISTENT_HANDLES)
@test.RequireShareCapabilities(smb2.SMB2_SHARE_CAP_CONTINUOUS_AVAILABILITY)
class Persistent(test.PikeTest):
    def setup(self):
        self.lease_key = array.array("B", map(random.randint, [0] * 16, [255] * 16))
        self.channel, self.tree = self.tree_connect()

    def create_persistent(self, prev_handle=0):
        """Helper to create a persistent handle, optionally reconnecting an old one"""
        handle = self.channel.create(
            self.tree,
            "persistent.txt",
            access=smb2.FILE_READ_DATA | smb2.FILE_WRITE_DATA | smb2.DELETE,
            share=SHARE_ALL,
            disposition=smb2.FILE_OPEN_IF,
            options=smb2.FILE_DELETE_ON_CLOSE,
            oplock_level=smb2.SMB2_OPLOCK_LEVEL_LEASE,
            lease_key=self.lease_key,
            lease_state=LEASE_RWH,
            durable=prev_handle,
            persistent=True,
        ).result()
        self.assertEqual(handle.durable_flags, smb2.SMB2_DHANDLE_FLAG_PERSISTENT)

        return handle

    # Opening a persistent handle grants persistent flag
    def test_create(self):
        handle = self.create_persistent()
        self.channel.close(handle)

    # Reconnecting a persistent handle after TCP disconnect works and preserves lease
    def test_reconnect(self):
        # Do a persistent open, drop connection, reconnect
        handle1 = self.create_persistent()
        self.channel.connection.close()
        self.channel, self.tree = self.tree_connect()

        handle2 = self.create_persistent(prev_handle=handle1)

        # Was the original lease state granted in the response?
        self.assertEqual(handle2.lease.lease_state, LEASE_RWH)

        self.channel.close(handle2)

    # Opening a file with a disconnected persistent handle fails with
    # STATUS_FILE_NOT_AVAILABLE
    def test_open_while_disconnected(self):
        # Do a persistent open, drop connection, reconnect
        handle1 = self.create_persistent()
        self.channel.connection.close()
        self.channel, self.tree = self.tree_connect()

        with self.assert_error(ntstatus.STATUS_FILE_NOT_AVAILABLE):
            # Perform a failing second open
            self.channel.create(
                self.tree,
                "persistent.txt",
                access=smb2.FILE_READ_DATA,
                share=SHARE_ALL,
                disposition=smb2.FILE_OPEN,
            ).result()

        # Clean up: Reconnect and close file (handle)
        self.channel.connection.close()
        self.channel, self.tree = self.tree_connect()
        handle2 = self.create_persistent(prev_handle=handle1)
        self.assertEqual(handle1.file_id, handle2.file_id)
        self.channel.close(handle2)

    def test_resiliency_same_timeout_reconnect_before_timeout(self):
        handle1 = self.create_persistent()
        self.assertTrue(handle1.is_persistent)
        timeout = handle1.durable_timeout
        a = self.channel.network_resiliency_request(handle1, timeout=timeout)
        self.assertTrue(handle1.is_resilient)
        self.channel.connection.close()

        #  sleeping time = resiliemcy < ca's default 120s
        time.sleep(min(handle1.durable_timeout / 1000.0, 1))
        self.channel, self.tree = self.tree_connect(client=pike.model.Client())

        # Invalidate handle from separate client
        with self.assert_error(ntstatus.STATUS_FILE_NOT_AVAILABLE):
            handle2 = self.channel.create(
                self.tree,
                "persistent.txt",
                access=smb2.FILE_READ_DATA,
                share=SHARE_ALL,
                disposition=pike.smb2.FILE_OPEN,
            ).result()

        self.channel.connection.close()

        self.channel, self.tree = self.tree_connect()
        handle3 = self.create_persistent(prev_handle=handle1)

        # handle1 and handle3 has the same file_id, and the same persistent status
        # it seams that resilience has not impacted the result of reconnect
        self.assertTrue(handle1.is_persistent)
        self.assertTrue(handle3.is_persistent)
        self.assertTrue(handle3.is_resilient)
        self.assertEqual(handle1.file_id, handle3.file_id)
        self.channel.close(handle3)

    def test_resiliency_reconnect_before_timeout(self):
        handle1 = self.create_persistent()
        self.assertTrue(handle1.is_persistent)
        timeout = 4000
        wait_time = timeout / 1000.0 * 0.5
        a = self.channel.network_resiliency_request(handle1, timeout=timeout)
        self.assertTrue(handle1.is_resilient)

        self.channel.connection.close()

        #  sleeping time < resiliemcy < ca's default 120s
        time.sleep(wait_time)

        self.channel, self.tree = self.tree_connect(client=pike.model.Client())
        with self.assert_error(ntstatus.STATUS_FILE_NOT_AVAILABLE):
            handle2 = self.channel.create(
                self.tree,
                "persistent.txt",
                access=smb2.FILE_READ_DATA,
                share=SHARE_ALL,
                disposition=pike.smb2.FILE_OPEN,
            ).result()

        self.channel.connection.close()

        self.channel, self.tree = self.tree_connect()
        handle3 = self.create_persistent(prev_handle=handle1)

        self.assertTrue(handle1.is_persistent)
        self.assertTrue(handle3.is_persistent)
        self.assertTrue(handle3.is_resilient)
        self.assertEqual(handle1.file_id, handle3.file_id)
        self.channel.close(handle3)

    def test_resiliency_reconnect_after_timeout(self):
        handle1 = self.create_persistent()
        self.assertTrue(handle1.is_persistent)
        timeout = 2000
        wait_time = timeout / 1000.0 * 1.1
        # ensure that the wait_time is _less_ than the granted timeout
        # to confirm that the resiliency request updates the timeout
        self.assertLess(wait_time, handle1.durable_timeout)
        a = self.channel.network_resiliency_request(handle1, timeout=timeout)
        self.assertTrue(handle1.is_resilient)
        self.channel.connection.close()

        # resiliemcy < sleeping time < ca's default 120s
        time.sleep(wait_time)

        self.channel, self.tree = self.tree_connect(client=pike.model.Client())

        # Because resiliency timeout, so another opener:handle2 break the persistent
        handle2 = self.channel.create(
            self.tree,
            "persistent.txt",
            access=smb2.FILE_READ_DATA,
            share=SHARE_ALL,
            disposition=pike.smb2.FILE_OPEN,
        )
        self.channel.connection.close()

        self.channel, self.tree = self.tree_connect()
        # Because resiliency timeout, other opener has broken the persistent, so reconnect would fail
        with self.assert_error(pike.ntstatus.STATUS_OBJECT_NAME_NOT_FOUND):
            handle3 = self.create_persistent(prev_handle=handle1)

    def test_resiliency_same_timeout_reconnect_after_timeout(self):
        handle1 = self.create_persistent()
        self.assertTrue(handle1.is_persistent)

        # set ca's timeout equals resiliency's timeout
        timeout = handle1.durable_timeout
        a = self.channel.network_resiliency_request(handle1, timeout=timeout)
        self.assertTrue(handle1.is_resilient)
        self.channel.connection.close()

        # resilience's timeout = ca's defaut 120s < sleeping time
        time.sleep(handle1.durable_timeout / 1000.0 + 1)

        # timeout, can't get object name
        self.channel, self.tree = self.tree_connect()
        with self.assert_error(pike.ntstatus.STATUS_OBJECT_NAME_NOT_FOUND):
            handle3 = self.create_persistent(prev_handle=handle1)

    def test_buffer_too_small(self):
        handle1 = self.create_persistent()
        self.assertTrue(handle1.is_persistent)

        timeout = 5
        with self.assert_error(pike.ntstatus.STATUS_BUFFER_TOO_SMALL):  # just for onefs
            smb_req = self.channel.request(obj=handle1.tree)
            ioctl_req = pike.smb2.IoctlRequest(smb_req)
            vni_req = InvalidNetworkResiliencyRequestRequest(ioctl_req)
            ioctl_req.file_id = handle1.file_id
            ioctl_req.flags = pike.smb2.SMB2_0_IOCTL_IS_FSCTL
            vni_req.Timeout = timeout
            vni_req.Reserved = 0
            a = self.channel.connection.transceive(ioctl_req.parent.parent)[0]

        self.channel.connection.close()

        self.channel, self.tree = self.tree_connect(client=pike.model.Client())
        with self.assert_error(ntstatus.STATUS_FILE_NOT_AVAILABLE):
            handle2 = self.channel.create(
                self.tree,
                "persistent.txt",
                access=smb2.FILE_READ_DATA,
                share=SHARE_ALL,
                disposition=pike.smb2.FILE_OPEN,
            ).result()

        self.channel.connection.close()

        self.channel, self.tree = self.tree_connect()

        # resiliency's buffer too small has not impacted persistent status
        handle3 = self.create_persistent(prev_handle=handle1)
        self.assertTrue(handle1.is_persistent)
        self.assertTrue(handle3.is_persistent)
        self.assertFalse(handle3.is_resilient)
        self.assertEqual(handle1.file_id, handle3.file_id)
        self.channel.close(handle3)



#
# Copyright (c) 2013-2020, Dell Inc. or its subsidiaries.
# All rights reserved.
# See file LICENSE for licensing information.
#
# Module Name:
#
#        digest.py
#
# Abstract:
#
#        Message digest and key derivation (for SMB2/3)
#
# Authors: Brian Koropoff (brian.koropoff@emc.com)
#

from __future__ import absolute_import
from __future__ import division
from builtins import range

import array

try:
    import Crypto.Hash.HMAC as HMAC
    import Crypto.Hash.SHA256 as SHA256
    import Crypto.Hash.SHA512 as SHA512
    import Crypto.Cipher.AES as AES
except ImportError:
    import Cryptodome.Hash.HMAC as HMAC
    import Cryptodome.Hash.SHA256 as SHA256
    import Cryptodome.Hash.SHA512 as SHA512
    import Cryptodome.Cipher.AES as AES

from . import core


def sha256_hmac(key, message):
    return array.array(
        "B",
        HMAC.new(
            key.tobytes(),
            message.tobytes(),
            SHA256,
        ).digest(),
    )


def aes128_cmac(key, message):
    aes = AES.new(key.tobytes(), mode=AES.MODE_ECB)

    def shiftleft(data):
        cin = 0
        cout = 0
        for i in reversed(range(0, len(data))):
            cout = (data[i] & 0x80) >> 7
            data[i] = ((data[i] << 1) | cin) & 0xFF
            cin = cout

        return cout

    def xor(data1, data2):
        for i in range(0, len(data1)):
            data1[i] ^= data2[i]

    def subkeys(key):
        zero = array.array("B", [0] * 16)
        rb = array.array("B", [0] * 15 + [0x87])

        key1 = array.array("B", aes.encrypt(zero.tobytes()))

        if shiftleft(key1):
            xor(key1, rb)

        key2 = array.array("B", key1)

        if shiftleft(key2):
            xor(key2, rb)

        return (key1, key2)

    message = array.array("B", message)
    mac = array.array("B", [0] * 16)
    scratch = array.array("B", [0] * 16)
    n = (len(message) + 16 - 1) // 16
    rem = len(message) % 16
    last_complete = n != 0 and rem == 0
    i = 0

    if n == 0:
        n = 1

    subkey1, subkey2 = subkeys(array.array("B", key))

    for i in range(0, n - 1):
        xor(mac, message[i * 16 : i * 16 + 16])
        mac = array.array("B", aes.encrypt(mac.tobytes()))

    if last_complete:
        scratch[0:16] = message[-16:]
        xor(scratch, subkey1)
    else:
        scratch[0:rem] = message[-rem:]
        scratch[rem] = 0x80
        xor(scratch, subkey2)

    xor(mac, scratch)
    mac = array.array("B", aes.encrypt(mac.tobytes()))

    return mac


def smb3_sha512(message):
    return array.array("B", SHA512.new(message.tobytes()).digest())


def derive_key(key, label, context):
    message = array.array("B")
    cur = core.Cursor(message, 0)

    cur.encode_uint32be(1)
    cur.encode_bytes(label)
    cur.encode_uint8be(0)
    cur.encode_uint8be(0)
    cur.encode_bytes(context)
    cur.encode_uint32be(len(key) * 8)

    return sha256_hmac(key, message)



#
# Copyright (c) 2020, Dell Inc. or its subsidiaries.
# All rights reserved.
# See file LICENSE for licensing information.
#
# Module Name:
#
#        test.py
#
# Abstract:
#
#        Test infrastructure
#
# Authors: Brian Koropoff (brian.koropoff@emc.com)
#

from __future__ import print_function
from __future__ import division
from builtins import object
from builtins import str
from future.utils import raise_from

import contextlib
import enum
import gc
import logging
import os
import unittest
import sys

import attr

import pike.model as model
import pike.ntstatus as ntstatus
import pike.smb2 as smb2


_NotSet = object()
_Required = object()


class MissingArgument(Exception):
    pass


class SequenceError(Exception):
    """Raised when calling TreeConnect helpers in the wrong state"""


class TestRequirementNotMet(Exception):
    """The test requires some dialect or capability that was not met"""


class DialectMissing(TestRequirementNotMet):
    pass


class CapabilityMissing(TestRequirementNotMet):
    pass


class ShareCapabilityMissing(TestRequirementNotMet):
    pass


class Options(enum.Enum):
    """
    Provide default Client and Test options sourced from the following
    Environment Variables.
    """

    PIKE_LOGLEVEL = "PIKE_LOGLEVEL"
    """Set the console log level by name (``DEBUG``, ``INFO``, etc)"""
    PIKE_TRACE = "PIKE_TRACE"
    """If 'yes' log full request/response payloads at DEBUG level"""
    PIKE_SERVER = "PIKE_SERVER"
    """SMB host name or IP address"""
    PIKE_PORT = "PIKE_PORT"
    """SMB port (default: {})""".format(model.default_port)
    PIKE_CREDS = "PIKE_CREDS"
    """
    Percent-delimited server credentials for NTLM authentication:
    ``DOMAIN\\User%Passwd``.

    If credentials are not specified and kerberos is available, attempt to use
    kerberos authentication.
    """
    PIKE_SHARE = "PIKE_SHARE"
    """Share name to connect to"""
    PIKE_SIGN = "PIKE_SIGN"
    """If 'yes' require packets to be signed"""
    PIKE_ENCRYPT = "PIKE_ENCRYPT"
    """If 'yes' require packets to be encrypted (SMB3-only)"""
    PIKE_MIN_DIALECT = "PIKE_MIN_DIALECT"
    """
    Minimum dialect to negotiate (``DIALECT_SMB2_1``, ``DIALECT_SMB3_0``, etc).

    See :py:class:`pike.smb2.Dialect`
    """
    PIKE_MAX_DIALECT = "PIKE_MAX_DIALECT"
    """
    Maximum dialect to negotiate (``DIALECT_SMB3_0``, ``DIALECT_SMB3_1_1``,
    etc).

    See :py:class:`pike.smb2.Dialect`
    """

    @classmethod
    def option(cls, name, default=None):
        if isinstance(name, cls):
            # convert from Enum type to str
            name = name.value
        value = os.environ.get(name, _NotSet)
        if value is _NotSet or len(value) == 0:
            if default is _Required:
                raise MissingArgument(
                    "Environment variable {!r} must be set".format(name)
                )
            value = default
        return value

    @classmethod
    def booloption(cls, name, default="no"):
        table = {
            "yes": True,
            "true": True,
            "1": True,
            "no": False,
            "false": False,
            "": False,
            "0": False,
        }
        return table[cls.option(name, default=default).lower()]

    @classmethod
    def smb2constoption(cls, name, default=None):
        return getattr(smb2, cls.option(name, "").upper(), default)

    @classmethod
    def loglevel(cls):
        return getattr(logging, cls.option(cls.PIKE_LOGLEVEL, default="NOTSET").upper())

    @classmethod
    def trace(cls):
        return cls.booloption(cls.PIKE_TRACE)

    @classmethod
    def server(cls):
        return cls.option(cls.PIKE_SERVER, default=_Required)

    @classmethod
    def port(cls):
        return int(cls.option(cls.PIKE_PORT, default=model.default_port))

    @classmethod
    def creds(cls):
        return cls.option(cls.PIKE_CREDS)

    @classmethod
    def share(cls):
        return cls.option(cls.PIKE_SHARE, "c$")

    @classmethod
    def signing(cls):
        return cls.booloption(cls.PIKE_SIGN)

    @classmethod
    def encryption(cls):
        return cls.booloption(cls.PIKE_ENCRYPT)

    @classmethod
    def min_dialect(cls):
        return cls.smb2constoption(cls.PIKE_MIN_DIALECT, default=0)

    @classmethod
    def max_dialect(cls):
        return cls.smb2constoption(cls.PIKE_MAX_DIALECT, default=float("inf"))


def default_client(signing=None):
    client = model.Client()
    min_dialect = Options.min_dialect()
    max_dialect = Options.max_dialect()
    client.restrict_dialects(min_dialect, max_dialect)
    if signing is None:
        signing = Options.signing()
    if signing:
        client.security_mode = (
            smb2.SMB2_NEGOTIATE_SIGNING_ENABLED | smb2.SMB2_NEGOTIATE_SIGNING_REQUIRED
        )
    return client


@attr.s
class TreeConnect(object):
    """
    Combines a :py:class:`~pike.model.Client`,
    :py:class:`~pike.model.Connection`, :py:class:`~pike.model.Channel`, and
    :py:class:`~pike.model.Tree` for simple access to an SMB share.

    May be used as a contextmanager: entering the context will establish the
    connection, session, and tree, and exiting the context will cleanly
    disconnect (if possible). Errors on disconnect are ignored.

    Any values not specified in the constructor will be drawn from the
    associated environment variables. At minimum the following should be
    specified explicitly or in the environment:

      * ``server`` (``PIKE_SERVER``) - server or hostname to connect to
      * ``creds`` (``PIKE_CREDS``) - percent-delimited credential string
      * ``share`` (``PIKE_SHARE``) - share name to connect to

    All other arguments are optional.

    The ``TreeConnect`` object can be used with the division operator, ``/``,
    which returns a :py:class:`~pike.path.PikePath` instance representing the
    path extended from the share root. The tree connection will be established
    before returning the :py:class:`~pike.path.PikePath` if is is not already
    connected.

    :vartype conn: pike.model.Connection
    :ivar conn: the Connection object, or None if not connected
    :vartype chan: pike.model.Channel
    :ivar chan: the Channel object, or None if no session is established
    :vartype tree: pike.model.Tree
    :ivar tree: the Tree object, or None if no tree is connected
    """

    _client = attr.ib(default=None)
    server = attr.ib(factory=Options.server)
    port = attr.ib(factory=Options.port)
    creds = attr.ib(factory=Options.creds)
    share = attr.ib(factory=Options.share)
    resume = attr.ib(default=None)
    signing = attr.ib(factory=Options.signing)
    encryption = attr.ib(factory=Options.encryption)
    require_dialect = attr.ib(default=None)
    require_capabilities = attr.ib(default=None)
    require_share_capabilities = attr.ib(default=None)
    conn = attr.ib(default=None, init=False)
    chan = attr.ib(default=None, init=False)
    tree = attr.ib(default=None, init=False)

    @require_dialect.validator
    def _require_dialect_validator(self, attribute, value):
        if value and len(value) != 2:
            raise TypeError(
                "require_dialect must be specified as a 2-tuple of (min_dialect, "
                "max_dialect), not {!r}".format(value)
            )

    @property
    def client(self):
        return self._client or default_client(signing=self.signing)

    def connect(self):
        """
        Establish a connection to the server and complete SMB2 NEGOTIATE.

        If a require_dialect or require_capability is specified to __init__, then
        an exception will be raised if the server does not support the required dialect
        or does not advertise the required capability.

        :return: connected pike.model.Connection
        """
        if self.conn and self.conn.connected:
            raise SequenceError(
                "Already connected: {!r}. Must call close() before reconnecting".format(
                    self.conn
                )
            )
        self.conn = self.client.connect(server=self.server, port=self.port).negotiate()
        negotiated_dialect = self.conn.negotiate_response.dialect_revision
        if self.require_dialect and (
            negotiated_dialect < self.require_dialect[0]
            or negotiated_dialect > self.require_dialect[1]
        ):
            self.close()
            raise DialectMissing("Dialect required: {}".format(self.require_dialect))

        capabilities = self.conn.negotiate_response.capabilities
        if self.require_capabilities and (
            capabilities & self.require_capabilities != self.require_capabilities
        ):
            self.close()
            raise CapabilityMissing(
                "Server does not support: %s "
                % str(self.require_capabilities & ~capabilities)
            )
        return self.conn

    def session_setup(self):
        """
        Establish a session on the connection and complete SMB2 SESSION_SETUP.

        If resume is specified to __init__, the new session will include the resumed
        session as previous_session_id.

        If encryption is specified to __init__ (or with PIKE_ENCRYPT environment var)
        the new session will encrypt data by default.

        :return: pike.model.Channel
        """
        if not self.conn or not self.conn.connected:
            raise SequenceError("Not connected. Must call connect() first")
        if self.chan:
            raise SequenceError(
                "Channel already established: {!r}. Must call close() before reconnecting".format(
                    self.chan
                )
            )
        self.chan = self.conn.session_setup(self.creds, resume=self.resume)
        if self.encryption:
            self.chan.session.encrypt_data = True
        return self.chan

    def tree_connect(self):
        """
        Establish a tree connection on the session and complete SMB2 TREE_CONNECT.

        If require_share_capability is specified to __init__, then
        an exception will be raised if the share does not does not advertise the
        required capability.

        :return: pike.model.Tree
        """
        if not self.chan:
            raise SequenceError(
                "Channel not established. Must call session_setup() first"
            )
        if self.tree:
            raise SequenceError(
                "Tree already connected: {!r}. Must call close() before reconnecting".format(
                    self.tree
                )
            )
        self.tree = self.chan.tree_connect(self.share)
        capabilities = self.tree.tree_connect_response.capabilities
        if self.require_share_capabilities and (
            capabilities & self.require_share_capabilities
            != self.require_share_capabilities
        ):
            self.close()
            raise ShareCapabilityMissing(
                "Share does not support: %s"
                % str(self.require_share_capabilities & ~capabilities)
            )
        return self.tree

    def __call__(self):
        """
        Perform all initialization steps (if needed). If the connection, channel, or
        tree is already established, this call is a no-op for those objects.

        :return: the established TreeConnect instance
        """
        if not self.conn:
            self.connect()
        if not self.chan:
            self.session_setup()
        if not self.tree:
            self.tree_connect()
        return self

    def close(self):
        """
        Perform de-initialization for all established objects. If any object has
        already been disconnected or set to None, then nothing will happen.

        For example, to perform SESSION_LOGOFF without first doing a TREE_DISCONNECT,
        simply set self.tree = None before calling close().

        Additionally, to disconnect without SESSION_LOGOFF, set self.chan = None.

        If the tree has already been disconnected or channel already closed, then
        errors related to cleaning up twice are suppressed.
        """
        session = None
        if self.chan and self.conn:
            session = self.conn.session(self.chan.session.session_id)
        if not self.conn or not self.conn.connected or not session:
            self.conn = self.chan = self.tree = None
            return
        if self.tree and self.chan:
            try:
                self.chan.tree_disconnect(self.tree)
            except model.ResponseError as err:
                if err.response.status == ntstatus.STATUS_USER_SESSION_DELETED:
                    self.chan = None
                elif err.response.status != ntstatus.STATUS_NETWORK_NAME_DELETED:
                    raise
            except EOFError:
                self.chan = None
            self.tree = None
        if self.chan:
            try:
                self.chan.logoff()
            except model.ResponseError as err:
                if err.response.status != ntstatus.STATUS_USER_SESSION_DELETED:
                    raise
            except EOFError:
                pass
            self.chan = self.tree = None
        if self.conn.connected:
            self.conn.close()
            self.conn = self.chan = self.tree = None

    def __enter__(self):
        """
        Context manager protocol. Establish connection, channel, and tree if not
        already established.

        :return: self
        """
        return self()

    def __exit__(self, exc_type, exc_val, exc_tb):
        """
        Context manager protocol. Calls close() to close any open objects.
        """
        self.close()

    def __del__(self):
        """
        Garbage collection. Calls close() to close any open objects.
        """
        self.close()

    def __iter__(self):
        """
        Compatibility shim to mirror the previous return value of PikeTest.tree_connect
        """
        yield self.chan
        yield self.tree

    def __getitem__(self, item):
        """
        Compatibility shim to mirror the previous return value of PikeTest.tree_connect
        """
        return tuple(self)[item]

    # proxy Path-like interface onto the tree
    @property
    def __fspath__(self):
        """
        proxy to Tree.__fspath__, establishing the session and tree if needed.
        """
        tree = self.tree or self().tree
        return tree.__fspath__

    @property
    def __truediv__(self):
        """
        proxy to Tree.__truediv__, establishing the session and tree if needed.
        """
        tree = self.tree or self().tree
        return tree.__truediv__

    if sys.version_info < (3,):
        __div__ = __truediv__


class PikeTest(unittest.TestCase):
    init_done = False

    @staticmethod
    def init_once():
        if not PikeTest.init_done:
            PikeTest.loglevel = Options.loglevel()
            PikeTest.handler = logging.StreamHandler()
            PikeTest.handler.setLevel(PikeTest.loglevel)
            PikeTest.handler.setFormatter(
                logging.Formatter("%(asctime)s:%(name)s:%(levelname)s: %(message)s")
            )
            PikeTest.logger = logging.getLogger("pike")
            PikeTest.logger.addHandler(PikeTest.handler)
            PikeTest.logger.setLevel(PikeTest.loglevel)
            model.trace = PikeTest.trace = Options.trace()
            PikeTest.init_done = True

    def __init__(self, *args, **kwargs):
        unittest.TestCase.__init__(self, *args, **kwargs)
        self.init_once()
        self._connections = []
        self.default_client = default_client()
        self.server = Options.server()
        self.port = Options.port()
        self.creds = Options.creds()
        self.share = Options.share()
        self.signing = Options.signing()
        self.encryption = Options.encryption()
        self.min_dialect = Options.min_dialect()
        self.max_dialect = Options.max_dialect()

    def debug(self, *args, **kwargs):
        self.logger.debug(*args, **kwargs)

    def info(self, *args, **kwargs):
        self.logger.info(*args, **kwargs)

    def warn(self, *args, **kwargs):
        self.logger.warn(*args, **kwargs)

    def error(self, *args, **kwargs):
        self.logger.error(*args, **kwargs)

    def critical(self, *args, **kwargs):
        self.logger.critical(*args, **kwargs)

    def set_client_dialect(self, min_dialect=None, max_dialect=None, client=None):
        if client is None:
            client = self.default_client
        client.restrict_dialects(min_dialect, max_dialect)

    def tree_connect(self, client=None, resume=None):
        tc = TreeConnect(
            client=client or self.default_client,
            server=self.server,
            port=self.port,
            creds=self.creds,
            share=self.share,
            resume=resume,
            encryption=self.encryption,
            require_dialect=self.required_dialect(),
            require_capabilities=self.required_capabilities(),
            require_share_capabilities=self.required_share_capabilities(),
        )
        try:
            tc()
        except TestRequirementNotMet as err:
            raise_from(unittest.SkipTest(str(err)), err)
        # save a reference to the TreeConnect object to avoid it being __del__'d
        self._connections.append(tc)
        return tc

    class _AssertErrorContext(object):
        pass

    @contextlib.contextmanager
    def assert_error(self, status):
        e = None
        o = PikeTest._AssertErrorContext()

        try:
            yield o
            raise self.failureException('No error raised when "%s" expected' % status)
        except model.ResponseError as err:
            o.response = err.response
            if err.response.status != status:
                raise_from(
                    self.failureException(
                        '"%s" raised when "%s" expected'
                        % (err.response.status, status),
                    ),
                    err,
                )

    def setUp(self):
        if self.loglevel != logging.NOTSET:
            print(file=sys.stderr)

        if hasattr(self, "setup"):
            self.setup()

    def tearDown(self):
        if hasattr(self, "teardown"):
            self.teardown()

        for conn in self._connections:
            conn.close()
        del self._connections[:]
        # Reference cycles are common in pike. so garbage collect aggressively
        gc.collect()

    def _get_decorator_attr(self, name, default):
        name = "__pike_test_" + name
        test_method = getattr(self, self._testMethodName)

        if hasattr(test_method, name):
            return getattr(test_method, name)
        elif hasattr(self.__class__, name):
            return getattr(self.__class__, name)
        else:
            return default

    def required_dialect(self):
        return self._get_decorator_attr("RequireDialect", (0, float("inf")))

    def required_capabilities(self):
        return self._get_decorator_attr("RequireCapabilities", 0)

    def required_share_capabilities(self):
        return self._get_decorator_attr("RequireShareCapabilities", 0)

    def assertBufferEqual(self, buf1, buf2):
        """
        Compare two sequences using a binary diff to efficiently determine
        the first offset where they differ
        """
        if len(buf1) != len(buf2):
            raise AssertionError("Buffers are not the same size")
        low = 0
        high = len(buf1)
        while high - low > 1:
            # XXX: consider usage of stdlib bisect module
            chunk_1 = (low, low + ((high - low) // 2))
            chunk_2 = (chunk_1[1], high)
            if buf1[chunk_1[0] : chunk_1[1]] != buf2[chunk_1[0] : chunk_1[1]]:
                low, high = chunk_1
            elif buf1[chunk_2[0] : chunk_2[1]] != buf2[chunk_2[0] : chunk_2[1]]:
                low, high = chunk_2
            else:
                break
        if high - low <= 1:
            raise AssertionError(
                "Block mismatch at byte {0}: "
                "{1} != {2}".format(low, buf1[low], buf2[low])
            )


class _Decorator(object):
    def __init__(self, value):
        self.value = value

    def __call__(self, thing):
        setattr(thing, "__pike_test_" + self.__class__.__name__, self.value)
        return thing


class _RangeDecorator(object):
    def __init__(self, minvalue=0, maxvalue=float("inf")):
        self.minvalue = minvalue
        self.maxvalue = maxvalue

    def __call__(self, thing):
        setattr(
            thing,
            "__pike_test_" + self.__class__.__name__,
            (self.minvalue, self.maxvalue),
        )
        return thing


class RequireDialect(_RangeDecorator):
    pass


class RequireCapabilities(_Decorator):
    pass


class RequireShareCapabilities(_Decorator):
    pass


class PikeTestSuite(unittest.TestSuite):
    """
    Custom test suite for easily patching in skip tests in downstream
    distributions of these test cases
    """

    skip_tests_reasons = {
        "test_to_be_skipped": "This test should be skipped",
    }

    @staticmethod
    def _raise_skip(reason):
        def inner(*args, **kwds):
            raise unittest.SkipTest(reason)

        return inner

    def addTest(self, test):
        testMethodName = getattr(test, "_testMethodName", None)
        if testMethodName in self.skip_tests_reasons:
            setattr(
                test,
                testMethodName,
                self._raise_skip(self.skip_tests_reasons[testMethodName]),
            )
        super(PikeTestSuite, self).addTest(test)


class SambaPikeTestSuite(PikeTestSuite):
    skip_tests_reasons = {
        # ERROR
        "test_resiliency_reconnect_after_timeout": "returns error against dperson/samba:d1a453d8123e462b0ad0ca8df51fb8ac0e5716b9",
        "test_resiliency_reconnect_before_timeout": "returns error against dperson/samba:d1a453d8123e462b0ad0ca8df51fb8ac0e5716b9",
        "test_resiliency_upgrade_reconnect_after_timeout": "returns error against dperson/samba:d1a453d8123e462b0ad0ca8df51fb8ac0e5716b9",
        "test_resiliency_upgrade_reconnect_before_timeout": "returns error against dperson/samba:d1a453d8123e462b0ad0ca8df51fb8ac0e5716b9",
        "test_bogus_resume_key": "returns error against dperson/samba:d1a453d8123e462b0ad0ca8df51fb8ac0e5716b9",
        "test_bogus_resume_key": "returns error against dperson/samba:d1a453d8123e462b0ad0ca8df51fb8ac0e5716b9",
        "test_other_resume_key": "returns error against dperson/samba:d1a453d8123e462b0ad0ca8df51fb8ac0e5716b9",
        "test_other_resume_key": "returns error against dperson/samba:d1a453d8123e462b0ad0ca8df51fb8ac0e5716b9",
        "test_ssc_in_compound_req": "returns error against dperson/samba:d1a453d8123e462b0ad0ca8df51fb8ac0e5716b9",
        "test_ssc_in_compound_req": "returns error against dperson/samba:d1a453d8123e462b0ad0ca8df51fb8ac0e5716b9",
        "test_set_sec_dacl": "returns error against dperson/samba:d1a453d8123e462b0ad0ca8df51fb8ac0e5716b9",
        "test_set_sec_dacl_append_ace": "returns error against dperson/samba:d1a453d8123e462b0ad0ca8df51fb8ac0e5716b9",
        "test_set_sec_dacl_new": "returns error against dperson/samba:d1a453d8123e462b0ad0ca8df51fb8ac0e5716b9",
        "test_set_sec_dacl_partial_copy": "returns error against dperson/samba:d1a453d8123e462b0ad0ca8df51fb8ac0e5716b9",
        "test_set_sec_group": "returns error against dperson/samba:d1a453d8123e462b0ad0ca8df51fb8ac0e5716b9",
        "test_set_sec_owner": "returns error against dperson/samba:d1a453d8123e462b0ad0ca8df51fb8ac0e5716b9",
        "test_set_sec_same": "returns error against dperson/samba:d1a453d8123e462b0ad0ca8df51fb8ac0e5716b9",
        "test_downlevel": "returns error against dperson/samba:d1a453d8123e462b0ad0ca8df51fb8ac0e5716b9",
        "test_downlevel": "returns error against dperson/samba:d1a453d8123e462b0ad0ca8df51fb8ac0e5716b9",
        "test_downlevel": "returns error against dperson/samba:d1a453d8123e462b0ad0ca8df51fb8ac0e5716b9",
        "test_treeconnect": "returns error against dperson/samba:d1a453d8123e462b0ad0ca8df51fb8ac0e5716b9",
        "test_write_none": "returns error against dperson/samba:d1a453d8123e462b0ad0ca8df51fb8ac0e5716b9",
        "test_write_none_lease": "returns error against dperson/samba:d1a453d8123e462b0ad0ca8df51fb8ac0e5716b9",
        "test_write_none_oplock": "returns error against dperson/samba:d1a453d8123e462b0ad0ca8df51fb8ac0e5716b9",
        "test_allow_zero_byte_write": "returns error against dperson/samba:d1a453d8123e462b0ad0ca8df51fb8ac0e5716b9",
        "test_set_get_reparse_point": "returns error against dperson/samba:d1a453d8123e462b0ad0ca8df51fb8ac0e5716b9",
        "test_symbolic_link_error_response": "returns error against dperson/samba:d1a453d8123e462b0ad0ca8df51fb8ac0e5716b9",
        "test_smb_3_0_encryption": "returns error against dperson/samba:d1a453d8123e462b0ad0ca8df51fb8ac0e5716b9",
        "test_smb_3_0_2_encryption": "returns error against dperson/samba:d1a453d8123e462b0ad0ca8df51fb8ac0e5716b9",
        "test_smb_3_1_1_compound": "returns error against dperson/samba:d1a453d8123e462b0ad0ca8df51fb8ac0e5716b9",
        "test_smb_3_1_1_encryption_ccm": "returns error against dperson/samba:d1a453d8123e462b0ad0ca8df51fb8ac0e5716b9",
        "test_smb_3_1_1_encryption_gcm": "returns error against dperson/samba:d1a453d8123e462b0ad0ca8df51fb8ac0e5716b9",
        # FAIL
        "test_resiliency_buffer_too_small": "returns fail against dperson/samba:d1a453d8123e462b0ad0ca8df51fb8ac0e5716b9",
        "test_resiliency_upgrade_buffer_too_small": "returns fail against dperson/samba:d1a453d8123e462b0ad0ca8df51fb8ac0e5716b9",
        "test_neg_dst_exc_brl": "returns fail against dperson/samba:d1a453d8123e462b0ad0ca8df51fb8ac0e5716b9",
        "test_neg_dst_is_a_dir": "returns fail against dperson/samba:d1a453d8123e462b0ad0ca8df51fb8ac0e5716b9",
        "test_neg_src_exc_brl": "returns fail against dperson/samba:d1a453d8123e462b0ad0ca8df51fb8ac0e5716b9",
        "test_smb3": "returns fail against dperson/samba:d1a453d8123e462b0ad0ca8df51fb8ac0e5716b9",
        "test_smb3_many_capabilities": "returns fail against dperson/samba:d1a453d8123e462b0ad0ca8df51fb8ac0e5716b9",
        "test_encryption_capabilities_both_prefer_gcm": "returns fail against dperson/samba:d1a453d8123e462b0ad0ca8df51fb8ac0e5716b9",
        "test_write_none_access": "returns fail against dperson/samba:d1a453d8123e462b0ad0ca8df51fb8ac0e5716b9",
        "test_query_interface_info_smb3": "returns fail against dperson/samba:d1a453d8123e462b0ad0ca8df51fb8ac0e5716b9",
        "test_create_failed_and_query": "returns fail against dperson/samba:d1a453d8123e462b0ad0ca8df51fb8ac0e5716b9",
        "test_deny_write": "returns fail against dperson/samba:d1a453d8123e462b0ad0ca8df51fb8ac0e5716b9",
        "test_async_lock": "returns fail against dperson/samba:d1a453d8123e462b0ad0ca8df51fb8ac0e5716b9",
        "test_async_write": "returns fail against dperson/samba:d1a453d8123e462b0ad0ca8df51fb8ac0e5716b9",
        "test_sequence_number_wrap": "returns fail against dperson/samba:d1a453d8123e462b0ad0ca8df51fb8ac0e5716b9",
        "test_qfid_diff_file": "returns fail against dperson/samba:7767ceb85af2e5254477357976feee49ca7eab3a",
        "test_qfid_same_file_seq_delete": "returns fail against dperson/samba:7767ceb85af2e5254477357976feee49ca7eab3a",
        "test_encryption_capabilities_both_prefer_ccm": "returns fail against dperson/samba:197dd6dac98274109c9c9c024f2bb1ebe2e075fa1ab901640a8fe94e875007d1",
    }


def suite(clz=PikeTestSuite):
    test_loader = unittest.TestLoader()
    test_loader.suiteClass = clz
    test_suite = test_loader.discover(
        os.path.abspath(os.path.dirname(__file__)), "*.py"
    )
    return test_suite


def samba_suite():
    return suite(SambaPikeTestSuite)


if __name__ == "__main__":
    test_runner = unittest.TextTestRunner()
    test_runner.run(suite())
